import sys
import ctypes

flags = sys.getdlopenflags()
sys.setdlopenflags(flags | ctypes.RTLD_GLOBAL)

# Dig past the extra python wrapping to the direct ScannerBit.so shared library interface level
import pyscannerbit.scan as sb
from pyscannerbit.ScannerBit.python import ScannerBit

# define likelihood, technically optional
def like(m):
    a = m["model1::x"]
    ScannerBit.print("my_param", 0.5) # can print custom parameters 

    return -a*a/2.0

# Version where ScannerBit interface is passed through via the wrapper
# (needed to allow multiple scans per python session)
def like2(scan,m):
    a = m["model1::x"]
    scan.print("my_param", 0.5) # can print custom parameters 
    return -a*a/2.0


# define prior, optional
def prior(vec, map):
    # tell ScannerBit that the hypergrid dimension is 1
    ScannerBit.ensure_size(vec, 1) # this needs to be the first line!

    map["model1::x"] = 5.0 - 10.0*vec[0]

# declare scan object
myscan = ScannerBit.scan(True)

settings = {
"Parameters": {
  "model1": {
    "x": None,
    }
  },
"Priors": {
  "x_prior": {
    "prior_type": 'flat',
    "parameters": ['model1::x'],
    "range": [1.0, 40.0],
    }
  },
"Printer": {
  "printer": "hdf5",
  "options": {
    "output_file": "results.hdf5",
    "group": "/",
    "delete_file_on_restart": "true",
    }
  },
"Scanner": {
  "scanners": {
    "twalk": {
      "plugin": "twalk",
      "like": "LogLike",
      "tolerance": 1.003,
      "kwalk_ratio": 0.9,
      "projection_dimension": 4
      }
    },
  "use_scanner": "twalk",
  },
"KeyValues": {
  "default_output_path": "pyscannerbit_run_data/",
  "likelihood": {
    "model_invalid_for_lnlike_below": -1e6
    }
  }
}

myscan.run(inifile=settings, lnlike={"LogLike": like}, prior=prior, restart=True)

# Try via wrapper to the above interface:
#sb._run_scan(settings, like2, prior)

#Try without prior
#sb._run_scan(settings, like2, "") # Fails!

# Try with dummy prior
def dummy_prior(vec, map):
    pass
 
#sb._run_scan(settings, like2, dummy_prior)







_default_options = {
"Printer": {
  "printer": "hdf5",
  "options": {
    "output_file": "results.hdf5",
    "group": "/",
    "delete_file_on_restart": "true",
    }
  },
"Scanner": {
  "scanners": {
    "diver": {
      "plugin": "diver",
      "like": "LogLike",
      "verbosity": 2,
      "NP": 1000,
      },
    "multinest": {
      "plugin": "multinest",
      "like":  "LogLike",
      "nlive": 1000,
      "tol": 0.5,
      },
    "polychord": {
      "plugin": "polychord",
      "like":  "LogLike",
      "nlive": 500,
      "tol": 0.5,
      },
    "twalk": {
      "plugin": "twalk",
      "sqrtR": 1.01, # This is a convergence criterion
      },
    "random": {
      "plugin": "random",
      "point_number": 10000,
      "like": "LogLike",
      },
    "toy_mcmc": {
      "like": "LogLike",
      "plugin": "toy_mcmc",
      "point_number": 10000,
      },
    },
  },
"KeyValues": {
  "default_output_path": "pyscannerbit_run_data/",
  "likelihood": {
    "model_invalid_for_lnlike_below": -1e6
    }
  }
}




# --- pyscannerbit import
import sys
import ctypes
flags = sys.getdlopenflags()
sys.setdlopenflags(flags | ctypes.RTLD_GLOBAL)
import pyscannerbit as sb
# ---

# ---  Tools for opening results and plotting
import h5py
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
# ---

# Choose YAML file to run
yaml = "test.yaml"

# Test function
def test(pars):
  x = pars["test_model::x"]
  y = pars["test_model::y"]
  z = pars["test_model::z"]
  #print "pars: ", x, y, z
  
  return -0.5*( (x-10)**2 + (y-15)**2 + (z+3)**2 ) # fitness function to maximise (log-likelihood)

#print sb.run_test.__doc__

skip_scan = True
if not skip_scan:
   sb.run_scan(yaml,test)

# Open the output file and examine scan results
f = h5py.File("runs/spartan/samples/results.hdf5",'r')
group = f["/spartan"]

# Inspect output datasets
print( [par for par in group] )

# Helper function to simplify parameter extraction
def get_entry(name):
   x = group[name]
   x_isvalid = np.array(group[name+"_isvalid"],dtype=np.bool)
   return x, x_isvalid

x, mx = get_entry("test_model::x")
y, my = get_entry("test_model::y")
z, mz = get_entry("test_model::z")
loglike, ml = get_entry("LogLike")

m = ml & mx & my & mz # mask for points with all valid parameters and likelihood

# Simple scatter plot
fig = plt.figure(figsize=(8,6))
ax = fig.add_subplot(111)
ax.scatter(x[m],y[m],facecolor='k',edgecolor='',alpha=0.5,s=2)
ax.set_xlabel("x")
ax.set_ylabel("y")
fig.savefig("test_output_xy_scatter.png")




import pyscannerbit.scan as sb
import pyscannerbit.defaults as defaults
import matplotlib.pyplot as plt
import copy

# Test function
def test_logl(x,y,z):
  #scan.print("x+y+z",x+y+z) # Save custom entry to output data
  return -0.5*( (x-10)**2 + (y-15)**2 + (z+3)**2 ) # fitness function to maximise (log-likelihood)

# Prior function; seems not to be optional like it should be.
def prior(vec, map):
  return map

# Override some scanner settings
# Easier to take the defaults and replace stuff rather than building from scratch
# (although if you do build it from scratch the defaults will be used to fill in
#  any gaps you leave)
settings = copy.deepcopy(defaults._default_options)
#settings["Scanner"]["scanners"]["multinest"] = {"tol": 0.5, "nlive": 500} # Configured for quick and dirty scan 
# Currently cannot use external scanners due to rpath issues. Use internal ones only for now.
#settings["Scanner"]["scanners"]["twalk"] = {"sqrtR": 1.05}
#settings["Scanner"]["scanners"]["random"] = {"point_number": 100}
settings["Scanner"]["scanners"]["toy_mcmc"] = {"point_number": 10}


# Create scan manager object
myscan = sb.Scan(test_logl, bounds=[[1., 40.]] * 3, prior_types=["flat", "flat", "log"], prior_func=prior, scanner="toy_mcmc", settings=settings)
print(myscan)
myscan.scan()

# Retrieve h5py group object, augmented with some helpful routines
hdf5 = myscan.get_hdf5()

# Variable names always match the ones used in "loglike"
print(hdf5.get_param_names())

# Best-fit parameters
print(hdf5.get_best_fit("x"))
print(hdf5.get_best_fit("y"))
print(hdf5.get_best_fit("z"))
print(hdf5.get_min_chi_squared())

# np.array of parameter
print(hdf5.get_param("z"))
print(hdf5.get_loglike())

# Plot pairs of parameters
hdf5.make_plot("x", "y")
hdf5.make_plot("x", "LogLike")

# Plot profile likelihood (requires an axis)
fig = plt.figure(figsize=(12,4))
ax = fig.add_subplot(121)
hdf5.plot_profile_likelihood(ax,"x","y")
ax = fig.add_subplot(122)
hdf5.plot_profile_likelihood(ax,"x","z")
plt.tight_layout()
fig.savefig("scan_object_test_logl.png")

# This is still an HDF5-like object (with the root being the group containing the datasets for this scan) 
# e.g., you can do
print(hdf5["default::x"])




"""Python interface to ScannerBit library, from GAMBIT project"""

# Need to do a little voodoo so that dynamically loaded shared libraries
# can dynamically load other shared libraries (the scanner plugins)
import os
import sys
import ctypes
flags = sys.getdlopenflags()
sys.setdlopenflags(flags | ctypes.RTLD_GLOBAL)

from functools import partial
import inspect
import copy
import h5py

# Just doing this will initialise MPI, and it will automatically
# call 'finalize' upon exit. So we need do nothing except import this.
from mpi4py import MPI

# DEBUG! Only used to check that YAML options looks reasonable
import yaml

# Need to tell ScannerBit where its config files are located
# We do this via a special environment variable
gambit_path = os.path.dirname(__file__)
print("Setting GAMBIT_RUN_DIR to:",gambit_path)
os.environ["GAMBIT_RUN_DIR"] = gambit_path

# Other python helper tools
from .defaults import _default_options
from .utils import _merge
from .hdf5_help import get_data, HDF5 
from .processify import processify

def _add_default_options(options):
   """Inspect user-supplied options, and fill in any missing
      bits with defaults"""
   _merge(options,_default_options)
   return options

# Better version of 'partial'. From 'funcy' library: https://github.com/Suor/funcy
def func_partial(func, *args, **kwargs):
    """A functools.partial alternative, which returns a real function.
       Can be used to construct methods."""
    return lambda *a, **kw: func(*(args + a), **dict(kwargs, **kw))

#@processify
def _run_scan(settings, loglike_func, prior_func):
   """Perform a scan. This function is decorated in such a 
      way that it runs in a new process. This is important
      because the GAMBIT plugins can only run once per
      process, because the shared libraries need to be reloaded
      to perform a second scan.
      """
   # Import functions from the ScannerBit.so library
   # Should also trigger the loading of the scanner plugin libraries
   from .ScannerBit.python import ScannerBit

   # Attach the ScannerBit object to the first argument of the wrapped likelihood function
   #wrapped_loglike = func_partial(loglike_func,scan=ScannerBit)

   # Create scan object
   myscan = ScannerBit.scan(True)
       
   # run scan
   # 'inifile' can be the name of a YAML file, or a dict.
   #myscan.run(inifile=settings, lnlike={"LogLike": wrapped_loglike}, prior=prior_func, restart=True)
   myscan.run(inifile=settings, lnlike={"LogLike": loglike_func}, prior=prior_func, restart=True)

class Scan:
    """Helper object for setting up and running a scan, and
       making some basic plots.
    """
    def __init__(self, function, prior_func=None, bounds=None, prior_types=None, kwargs=None, scanner=None,
      settings={}, model_name=None, output_path=None, fargs=None):
        """
        function - Python function to be scanned
        prior_func - User-define prior transformation function (optional)
        bounds - list of ranges of parameter values (function arguments) to scan,
                 or mean/std-dev in case of normal prior.
        prior_types - list of priors to use for scanning in each dimension (e.g. flat/log).
                      If None then the user is expected to do the inverse transform from
                      a unit hypercube to their parameter space themselves.
        scanner - Scanning algorithm to use
        f_args - List of names of function argments to use (if None these are
         inferred from the signature of 'function')
        """
        self.function = function
        self.prior_func = prior_func
        self.ScannerBit = None # To contain ScannerBit interface module when run begins

        # Determine parameter names, either automatically or from 'fargs' argument
        if fargs is None:
            signature = inspect.getargspec(self.function)
            # First argument needs to be the 'scan' object, to allow access to printers etc. Skip it in determination of parameter names.
            #self._argument_names = signature.args[1:]
            self._argument_names = signature.args
        else:
            self._argument_names = fargs

        # Determine prior to be used
        self.bounds = bounds if bounds else [(0,1)] * len(self._argument_names)
        self.prior_types = prior_types if prior_types else ["flat"] * len(self._argument_names)
        self.scanner = scanner
        self.settings = _add_default_options(copy.deepcopy(settings))
        self.kwargs = kwargs

        print("self._argument_names:", self._argument_names)
        print("self.bounds:", self.bounds)
        assert len(self._argument_names) == len(self.bounds)
 
        if model_name is None:
           if "Parameters" in self.settings:
              #print(self.settings["Parameters"])
              self._model_name = list(self.settings["Parameters"].keys())[0]
           else:
              self._model_name = "default"
        self._wrapped_function = self._wrap_function()
        self._scanned = False

        # Make up a name for the run, if user didn't provide one
        if output_path is None:
            self.settings["KeyValues"]["default_output_path"] += "{0}_scan".format(self.scanner)
        else:
            self.settings["KeyValues"]["default_output_path"] = output_path

        self._process_settings()
        self.loglike_par = "LogLike"
        self.scanner = self.settings["Scanner"]["use_scanner"] # Might have been None and then filled by _process_settings with defaults
        if self.scanner == "multinest" \
        or self.scanner == "polychord":
            self.posterior_par = "Posterior"
        else:
            self.posterior_par = None

    def _get_hdf5_group(self):
        """
        """
        assert self.settings["Printer"]["printer"] == "hdf5"
        file_name = self.settings["Printer"]["options"]["output_file"]
        group_name = self.settings["Printer"]["options"]["group"]
        DIR = self.settings["KeyValues"]["default_output_path"]
        fullpath = "{}/samples/{}".format(DIR, file_name)
        f = h5py.File(fullpath,'r')
        g = f[group_name]
        return g

    def _process_settings(self):
        """Copy 'pythonic' options into the settings dictionary
           so that they can be automaticall converted into YAML 
           by the API, which ScannerBit can then read"""
        if(self.scanner is not None):
            self.settings["Scanner"]["use_scanner"] = self.scanner
        if("Parameters" not in self.settings):
            self.settings["Parameters"] = dict()
            self.settings["Parameters"][self._model_name] = dict()
            for n in self._argument_names:
                # Just add parameter names. We will write the prior settings out fully
                # in the Priors section next
                self.settings["Parameters"][self._model_name][n] = None
        if("Priors" not in self.settings):
            self.settings["Priors"] = dict()
            if (self.prior_types is not None) \
               and (self.bounds is not None):
                for n, b, t in zip(self._argument_names, self.bounds, self.prior_types):
                    prior_setup = {'prior_type': t, 'parameters': ["{0}::{1}".format(self._model_name,n)]}
                    if t is 'gaussian' or t is 'cauchy':
                       prior_setup['mean'] = [b[0]]
                       prior_setup['cov']  = [b[1]] # Assume this is the VARIANCE, not standard deviation
                    else:
                       prior_setup['range'] = b
                    self.settings["Priors"]["{0}_prior".format(n)] = prior_setup
            else:
                raise ValueError("No prior settings found! These need to be either supplied in simplified form via the 'bounds' and 'prior_types' arguments, or else supplied in long form (following the GAMBIT YAML format) in the 'settings' dictionary under the 'Priors' key (or under the 'Parameters' key in the short-cut format)")
        #print(self.settings)
        #print("Scan settings:")
        #print("==============")
        #print(yaml.dump(self.settings, default_flow_style=False))

    def _wrap_function(self):
        """
        """
        def wrapped_function(par_dict):
            """
            """
            print("par_dict:", par_dict)
            arguments = [par_dict["{}::{}".format(self._model_name, n)]
              for n in self._argument_names]
            return self.function(*arguments, **(self.kwargs or {}))

        return wrapped_function

    def scan(self):
       """Perform a scan. This runs a function that is decorated in such a 
       way that it runs in a new process. This is important
       because the GAMBIT plugins can only run once per
       process, because the shared libraries need to be reloaded
       to perform a second scan.

       Downside is that all arguments must be pickle-able.
       """
       _run_scan(self.settings, self._wrapped_function, self.prior_func)
       MPI.COMM_WORLD.Barrier()
       rank = MPI.COMM_WORLD.Get_rank()
       print("Rank {0} passed scan end barrier!".format(rank))
       self._scanned = True

    def get_hdf5(self):
        try:
            g = HDF5(self._get_hdf5_group().id,model=self._model_name,
                loglike=self.loglike_par, posterior=self.posterior_par)
        except:
            if(self._scanned):
                raise IOError("Failed to open HDF5 output of scan!")
            else:
                raise IOError("Failed to open HDF5 output of scan, however we did not perform a scan just now. The output will only exist if you have previously run this scan. Please check that you did this!")
        return g



import pyscannerbit.scan as sb
import matplotlib.pyplot as plt

# Test function
def test_logl(x,y,z):
  return -0.5*( (x-10)**2 + (y-15)**2 + (z+3)**2 ) # fitness function to maximise (log-likelihood)

# Override some scanner settings (this is a little ugly for now)
# Can make it a little nicer with defaultdict and some recursion
from collections import defaultdict
def rec_dd():
    return defaultdict(rec_dd)
settings = rec_dd() # Uses a new dict as default value when accessing non-existant keys
settings["Scanner"]["scanners"]["multinest"] = {"tol": 0.5, "nlive": 500} # Configured for quick and dirty scan 

# Create scan manager object
myscan = sb.Scan(test_logl, bounds=[[1., 40.]] * 3, prior_types=["log", "flat", "log"], scanner="multinest", settings=settings)
myscan.scan()

# Retrieve h5py group object, augmented with some helpful routines
hdf5 = myscan.get_hdf5()

# Variable names always match the ones used in "loglike"
print(hdf5.get_param_names())

# Best-fit parameters
print(hdf5.get_best_fit("x"))
print(hdf5.get_best_fit("y"))
print(hdf5.get_best_fit("z"))
print(hdf5.get_min_chi_squared())

# np.array of parameter
print(hdf5.get_param("z"))
print(hdf5.get_loglike())

# Plot pairs of parameters
hdf5.make_plot("x", "y")
hdf5.make_plot("x", "LogLike")

# Plot profile likelihood (requires an axis)
fig = plt.figure(figsize=(12,4))
ax = fig.add_subplot(121)
hdf5.plot_profile_likelihood(ax,"x","y")
ax = fig.add_subplot(122)
hdf5.plot_profile_likelihood(ax,"x","z")
plt.tight_layout()
fig.savefig("scan_object_test_logl.png")

# This is still an HDF5-like object (with the root being the group containing the datasets for this scan) 
# e.g., you can do
print(hdf5["default::x"])




import import_yaml

template = """
Parameters:
  python_model: None

Priors:

Printer:

  printer: hdf5
  options:
    output_file: "results.hdf5"
    group: "/python"
    delete_file_on_restart: true

Scanner:

  use_scanner: None

  scanners:

    de:
      plugin: diver
      like: LogLike
      NP: 1000

    multinest:
      plugin: multinest
      like:  LogLike
      nlive: 1000
      tol: 0.1

    mcmc:
      plugin: great
      like: LogLike
      nTrialLists: 5
      nTrials: 10000

    twalk:
      plugin: twalk

Logger:
  redirection:
    [Default]      : "default.log"
    [Scanner]      : "Scanner.log"

KeyValues:
  likelihood:
    model_invalid_for_lnlike_below: -1e6
"""

settings = import_yaml.load(template)



from sb import Scan, HDF5
from default import settings

# Make a function to investigate

def loglike(x, y, z, extra=False):
  if extra:
      x += 1.
  return -0.5 * ((x - 10.)**2 + (y - 15.)**2 + (z - 30.)**2)
  
 
# Tweak default settings (optional)
 
settings["Scanner"]["scanners"]["de"]["NP"] = 100

# Scan over 1 to 40 for each parameter with "de" scanner, with a combination
# of log and flat priors etc.
  
scan = Scan(loglike, [[1., 40.]] * 3, prior_types=["log", "flat", "log"], kwargs={'extra': False}, scanner="multinest", settings=settings)
scan.scan()

# Extract HDF5 object and look at it

hdf5 = scan.get_hdf5()

# Variable names always match the ones used in "loglike"

print(hdf5.get_param_names())

# Best-fit parameters
print(hdf5.get_best_fit("x"))
print(hdf5.get_best_fit("y"))
print(hdf5.get_best_fit("z"))
print(hdf5.get_min_chi_squared())

# np.array of parameter
print(hdf5.get_param("z"))
print(hdf5.get_loglike())

# Plot pairs of parameters
hdf5.make_plot("x", "y")
hdf5.make_plot("x", "LogLike")

# This is still an HDF5-like object, e.g., you can do
print(hdf5["/python"]["python_model::x"])
